#
#
#


__author__ = "Randall J. Radmer"
__version__ = "1.0"
__doc__ = """The SimplePdb class reads, parses, and write simple PDB files"""

import defaultNastParameters as p

class SimplePdb(object):
    def __init__(self,
                 pdbFilename=None,
                 pdbString=None,
                 sequenceFilename=None,
                 sequenceString=None,
                 centerAtomNames=p.centerAtomNames):
        super(SimplePdb, self).__init__()

        self.centerAtomNames=centerAtomNames

        self.pdbFilename=None
        self.pdbString=None
        self.sequenceFilename=None
        self.sequenceString=None
        self.data=[]
        self.chainBreaks=[]
        self.originalPdbLines=None

        if pdbFilename:
            self.pdbFilename=pdbFilename
            f=open(pdbFilename)
            self.pdbString=f.read()
            f.close()
        elif pdbString:
            self.pdbString=pdbString
        elif sequenceFilename: 
            self.sequenceFilename=sequenceFilename
            f=open(sequenceFilename)
            self.sequenceString=f.read()
            f.close()
        elif sequenceString:
            self.sequenceString=sequenceString

        if self.pdbString:
            self.originalPdbLines=[]
            for line in self.pdbString.splitlines():
                if line.find("END")==0: break
                if line.find("TER")==0:
                    self.chainBreaks.append(len(self.data))
                if len(line)<53: continue
                if ( (line[0:4]=='ATOM' or line[0:6]=='HETATM') and
                     line[12:16].strip() in centerAtomNames):
                    self.data.append({})
                    self.data[-1]['atomNum'] = int(line[6:11])
                    self.data[-1]['atomName'] = line[12:16].strip()
                    self.data[-1]['resName'] = line[17:20].strip()
                    self.data[-1]['resNum'] = int(line[22:26])
                    self.data[-1]['x'] = float(line[30:38])/10.0
                    self.data[-1]['y'] = float(line[38:46])/10.0
                    self.data[-1]['z'] = float(line[46:54])/10.0
                    self.originalPdbLines.append(line)
            self.numParticles=len(self.data)
        elif self.sequenceString:
            import re, math
            count=0
            for residueName in re.split("[^a-zA-Z0-9_\*']+", self.sequenceString):
                if len(residueName)>0:
                    if residueName=='*':
                        self.chainBreaks.append(len(self.data))
                    else:
                        self.data.append({})
                        self.data[-1]['atomNum'] = count+1
                        self.data[-1]['atomName'] = p.centerAtomNames[0]
                        self.data[-1]['resName'] = residueName[:3]
                        self.data[-1]['resNum'] = count+1
                        count+=1
            self.numParticles=len(self.data)
            circleRadius=(p.bondR_bb*self.numParticles)/(2*math.pi)
            for i in range(self.numParticles):
                self.data[i]['x'] = circleRadius*math.cos(2*math.pi*i/float(self.numParticles))
                self.data[i]['y'] = circleRadius*math.sin(2*math.pi*i/float(self.numParticles))
                self.data[i]['z'] = 0.0

    def setPositions(self, positions):
        for i in len(self.data):
            self.data[i]['x']=positions[i][0]
            self.data[i]['y']=positions[i][1]
            self.data[i]['z']=positions[i][2]


    def getPositions(self):
        positions=[]
        for i in range(len(self.data)):
            positions.append( (self.data[i]['x'], 
                               self.data[i]['y'], 
                               self.data[i]['z']) )
        return positions

    def getPdb(self, positions=[], useOriginalPdb=False, verbose=False):
        """Make PDB format string"""
        pdbString=''
        if useOriginalPdb:
            if len(positions)==self.numParticles:
                for i in range(self.numParticles):
                    if i in self.chainBreaks: 
                        pdbString = '%sTER\n' % pdbString
                    line=self.originalPdbLines[i]
                    pdbString = '%s%s%8.2f%8.2f%8.2f\n' \
                               % (pdbString, line[:30],
                                  10*positions[i][0],
                                  10*positions[i][1],
                                  10*positions[i][2])
            else:
                for i in range(self.numParticles):
                    if i in self.chainBreaks: 
                        pdbString = '%sTER\n' % pdbString
                    line=self.originalPdbLines[i]
                    pdbString = '%s%s%8.2f%8.2f%8.2f\n' \
                               % (pdbString, line[:30],
                                  10*self.data[i]['x'],
                                  10*self.data[i]['y'],
                                  10*self.data[i]['z'])
        else:
            for i in range(self.numParticles):
                if verbose:
                    line=self.originalPdbLines[i]
                    pdbString = '%sREMARK%s\n' % (pdbString, line[6:])
                if i in self.chainBreaks: 
                    pdbString = '%sTER\n' % pdbString
                pdbString = '%sATOM   %4d  %3s %3s  %4d    ' \
                           % (pdbString,
                              self.data[i]['atomNum'],
                              self.data[i]['atomName'],
                              self.data[i]['resName'],
                              self.data[i]['resNum'])
                if len(positions)==self.numParticles:
                    pdbString = '%s%8.2f%8.2f%8.2f\n' \
                               % (pdbString,
                                  10*positions[i][0],
                                  10*positions[i][1],
                                  10*positions[i][2])
                else:
                    pdbString = '%s%8.2f%8.2f%8.2f\n' \
                               % (pdbString,
                                  10*self.data[i]['x'],
                                  10*self.data[i]['y'],
                                  10*self.data[i]['z'])

        return pdbString



    def getPsf(self, mass=p.mass):
        """Make X-PLOR format PSF string"""
        s='PSF\n\n'
        s+='       1 !NTITLE\n'
        s+=' REMARKS generated by nast.py\n\n'
    
        s+='%8d !NATOM\n' % self.numParticles
        for i in range(self.numParticles):
            s+= '%8d MAIN %-4d %3s %3s %17.6f %14.6f  0\n' \
               % (i+1, i+1, self.data[i]['resName'],
                  self.data[i]['atomName'], 0.0, mass)
    
        sBlock=''
        sLine=''
        nBonds=0
        for i in range(self.numParticles-1):
            if i+1 in self.chainBreaks: continue
            sLine+=' %7d %7d' % (i+1, i+2)
            nBonds+=1
            if 8<=len(sLine.split()):
                sBlock+=sLine+'\n'
                sLine=''
        if 0<len(sLine.split()):
            sBlock+=sLine+'\n'
        s+='\n%8d !NBOND: bonds\n' % nBonds
        s+=sBlock
    
        sBlock=''
        sLine=''
        nAngles=0
        for i in range(self.numParticles-2):
            if i+1 in self.chainBreaks: continue
            if i+2 in self.chainBreaks: continue
            sLine+=' %7d %7d %7d' % (i+1, i+2, i+3)
            nAngles+=1
            if 9<=len(sLine.split()):
                sBlock+=sLine+'\n'
                sLine=''
        if 0<len(sLine.split()):
            sBlock+=sLine+'\n'
        s+='\n%8d !NTHETA: angles\n' % nAngles
        s+=sBlock
    
        sBlock=''
        sLine=''
        nDihedrals=0
        for i in range(self.numParticles-3):
            if i+1 in self.chainBreaks: continue
            if i+2 in self.chainBreaks: continue
            if i+3 in self.chainBreaks: continue
            sLine+=' %7d %7d %7d %7d' % (i+1, i+2, i+3, i+4)
            nDihedrals+=1
            if 8<=len(sLine.split()):
                sBlock+=sLine+'\n'
                sLine=''
        if 0<len(sLine.split()):
            sBlock+=sLine+'\n'
        s+='\n%8d !NPHI: dihedrals\n' % nDihedrals
        s+=sBlock
    
        s+='\n%8d !NIMPHI: impropers\n' % 0
    
        s+='\n\n'
        return s



def parseCommandLine():
    opts, args_proper = getopt.getopt(sys.argv[1:], 'ho:p:')
    pdbOutFilename=None
    psfOutFilename=None
    for option, parameter in opts:
        if option=='-h': usageError()
        if option=='-o': pdbOutFilename=parameter
        if option=='-p': psfOutFilename=parameter
    return (args_proper, pdbOutFilename, psfOutFilename)

def runTest():
    args_proper, pdbOutFilename, psfOutFilename = parseCommandLine()
    try:
        pdbFilename = args_proper[0]
    except IndexError:
        usageError()
    pdb=SimplePdb(pdbFilename=pdbFilename)
    sys.stdout.write("PDB Filename: %s\n" % pdb.pdbFilename)
    sys.stdout.write("Number of Particles: %d\n" % pdb.numParticles)

    if pdbOutFilename:
        sys.stdout.write("New PDB Filename: %s\n" % pdbOutFilename)
        fOut=open(pdbOutFilename, 'w')
        fOut.write(pdb.getPdb())
        fOut.close()

    if psfOutFilename:
        sys.stdout.write("New PSF Filename: %s\n" % psfOutFilename)
        fOut=open(psfOutFilename, 'w')
        fOut.write(pdb.getPsf())
        fOut.close()


    sys.stdout.flush()
    return

def usageError():
    sys.stdout.write('usage: %s pdbInFilename [-o pdbOutFilename] [-p psfOutFilename]\n'
         % os.path.basename(sys.argv[0]))
    sys.exit(1)

if __name__=='__main__':
    import os, sys, getopt
    runTest()

